/*
  Hatari - main.c

  This file is distributed under the GNU Public License, version 2 or at
  your option any later version. Read the file gpl.txt for details.

  Main initialization and event handling routines.
*/

#include <time.h>
#include <signal.h>
#include <sys/time.h>
#include <unistd.h>
#include <signal.h>
#include <fcntl.h>

#include <stdint.h>
#include <stdio.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <errno.h>

#include <SDL.h>

#include "main.h"
#include "configuration.h"
#include "dialog.h"
#include "audio.h"
#include "file.h"
#include "../m68000.h"
#include "hostcall.h"
#include "input.h"
#include "keymap.h"
#include "misc.h"
#include "screen.h"
#include "sdlgui.h"
#include "shortcut.h"


#define FORCE_WORKING_DIR                 /* Set default directory to cwd */

#define GP2X_BUTTON_UP              (0)
#define GP2X_BUTTON_DOWN            (4)
#define GP2X_BUTTON_LEFT            (2)
#define GP2X_BUTTON_RIGHT           (6)
#define GP2X_BUTTON_UPLEFT          (1)
#define GP2X_BUTTON_UPRIGHT         (7)
#define GP2X_BUTTON_DOWNLEFT        (3)
#define GP2X_BUTTON_DOWNRIGHT       (5)
#define GP2X_BUTTON_CLICK           (18)
#define GP2X_BUTTON_A               (12)
#define GP2X_BUTTON_B               (13)
#define GP2X_BUTTON_Y               (14)
#define GP2X_BUTTON_X               (15)
#define GP2X_BUTTON_L               (10)
#define GP2X_BUTTON_R               (11)
#define GP2X_BUTTON_START           (8)
#define GP2X_BUTTON_SELECT          (9)
#define GP2X_BUTTON_VOLUP           (16)
#define GP2X_BUTTON_VOLDOWN         (17)

BOOL bQuitProgram=FALSE;                  /* Flag to quit program cleanly */
BOOL bUseFullscreen=FALSE;
BOOL bEmulationActive=TRUE;               /* Run emulation when started */
BOOL bAppActive = FALSE;
char szBootDiscImage[MAX_FILENAME_LENGTH] = { "" };

char szWorkingDir[MAX_FILENAME_LENGTH] = { "" };
char szCurrentDir[MAX_FILENAME_LENGTH] = { "" };

int ratonx=1,ratony=1,incraton=2;
int pu=0,pd=0,pl=0,pr=0,raton=0;
int bs=0,br=0,bl=0;
int savedrx=0, savedry=0;

#define DMODE_CURSOR 0
#define DMODE_FLIGHT 1
#define DMODE_MAP 2

int dmode = DMODE_CURSOR;

extern const unsigned char * mouse_bmp;
extern const unsigned char cursor_bmp[256];
extern const unsigned char stick_bmp[256];
extern const unsigned char world_bmp[256];

typedef struct {
    uint16_t               pressure;
    uint16_t              x;
    uint16_t               y;
    uint16_t               pad;
    struct timeval  stamp;
} TS_EVENT;



TS_EVENT ts_event;
//int fd;						// Buffer pa lectura de tactil
static int tactilx,tactily;
char autocalib=1;
static int calibX=0, calibY=0;
char tsrelease=0;


/*-----------------------------------------------------------------------*/
/*
  Error handler
*/
void Main_SysError(char *Error,char *Title)
{
  fprintf(stderr,"%s : %s\n",Title,Error);
}


/*-----------------------------------------------------------------------*/
/*
  Bring up message(handles full-screen as well as Window)
*/
int Main_Message(char *lpText, char *lpCaption/*,unsigned int uType*/)
{
  int Ret=0;

  /* Show message */
  fprintf(stderr,"%s: %s\n", lpCaption, lpText);

  return(Ret);
}


/*-----------------------------------------------------------------------*/
/*
  Pause emulation, stop sound
*/
void Main_PauseEmulation(void)
{
  if( bEmulationActive )
  {
    Audio_EnableAudio(FALSE);
    bEmulationActive = FALSE;
  }
}

/*-----------------------------------------------------------------------*/
/*
  Start emulation
*/
void Main_UnPauseEmulation(void)
{
  if( !bEmulationActive )
  {
    Audio_EnableAudio(ConfigureParams.Sound.bEnableSound);
    bFullScreenHold = FALSE;      /* Release hold  */

    bEmulationActive = TRUE;
  }
}

/* ----------------------------------------------------------------------- */
/*
  Message handler
  Here we process the SDL events (keyboard, mouse, ...) and map it to
  Atari IKBD events.
*/
void Main_EventHandler()
{
  SDL_Event event;

  if (pr)
  {
    if ((!raton)&&(ratonx<320))
    {
        ratonx += incraton;
        input.abs_x =ratonx;
    }
    input.motion_x +=incraton;
  }

  if (pl)
  {
    if ((!raton)&&(ratonx>0))
    {
        ratonx -= incraton;
        input.abs_x =ratonx;
    }
    input.motion_x -=incraton;
  }

  if (pu)
  {
    if ((!raton)&&(ratony>0))
    {
        ratony -= incraton;
        input.abs_y =ratony;
    }
    input.motion_y -=incraton;
  }

   if (pd)
   {
     if ((!raton)&&(ratony<200))
     {
        ratony += incraton;
        input.abs_y =ratony;
     }
     input.motion_y +=incraton;
   }


// Nuevo codigo de raton
#if 0
	if (SDL_GetMouseState(&tactilx, &tactily)&SDL_BUTTON(1)){

		if (tactily>200)
			tactily=200;

		if (tactily<0)
			tactily=0;

		if (tactilx>320)
			tactilx=320;

		if (tactilx<0)
			tactilx=0;
		

		ratonx=tactilx;
		ratony=tactily;

		input.abs_x =ratonx;
		input.abs_y =ratony;

                Input_MousePress (SDL_BUTTON_LEFT);
		tsrelease=1;

	}else{
		if (tsrelease==1){
		  Input_MouseRelease (SDL_BUTTON_LEFT);
		  tsrelease=0;
		}
	}
#endif	

/*****
 R + L + SELECT = QUIT
 SELECT = swap between "Mouse mode" and "Flight mode" and "Star Map Mode"

 Controls in Mouse Mode
 ======================
 Dpad = move mouse
 A = Left Click
 B = Right Click
 Y = Hyperspace to Destination (F8)
 X = Eject from ship (if you have an escape pod) (x)
 R = use radar mapper (r)
 L = hyperspace cloud analyser (???)

 Double click on pause icon = options (for adjusting detail level, saving and loading games etc)

 Controls in Flight mode
 =======================
 Dpad = yaw / pitch
 L = Decrease Thrust (RSHIFT)
 R = Increase Thrust (RETURN)
 A = Fire Laser (SPACE)
 B = Status Ship (F3)
 Y = ECM (e)
 X = Change View (F1)
 START = Motor (F7)
 ? = with military camera takes photograph, with MB4 mining machine deploys MB4
 ? = Fire Missile when targeted
 ? = Drop Energy Bomb

*/
  while (SDL_PollEvent (&event))
   switch( event.type )
   {


    case SDL_QUIT:
       bQuitProgram = TRUE;
       SDL_Quit ();
       exit (0);
       break;




    case SDL_KEYDOWN:
	switch(event.key.keysym.sym)
        {
		case SDLK_RIGHT:
		        if (dmode == DMODE_MAP)
             		    Keymap_KeyDown(&event.key.keysym);
			else
             		    pr=1;

	     		break;

		case SDLK_LEFT:
		        if (dmode == DMODE_MAP)
             		    Keymap_KeyDown(&event.key.keysym);
                        else
             		    pl=1;
	     		break;

		case SDLK_UP:
		        if (dmode == DMODE_MAP)
             		    Keymap_KeyDown(&event.key.keysym);
			else
                            pu=1;
	     		break;

		case SDLK_DOWN:
		        if (dmode == DMODE_MAP)
             		    Keymap_KeyDown(&event.key.keysym);
			else
                            pd=1;

	     		break;

           	case SDLK_LCTRL: //A
             		    Input_MousePress (SDL_BUTTON_LEFT);
             		break;

		case SDLK_LALT:  //B
		        if(dmode == DMODE_FLIGHT)
                        {
             		    event.key.keysym.sym=SDLK_F3; //status ship
             		    Keymap_KeyDown(&event.key.keysym);
                        }
                        else
                        {
                            raton = 1;
             		    Input_MousePress (SDL_BUTTON_RIGHT);
                        }

			break;

                case SDLK_TAB: //L
                        bl=1;
		        if(dmode == DMODE_FLIGHT)
             		    event.key.keysym.sym=SDLK_RSHIFT; //- throttle
			else
             		    event.key.keysym.sym=SDLK_h; // hyperspace cloud analyser ???

            		Keymap_KeyDown(&event.key.keysym);
                        break;

		case SDLK_BACKSPACE: //R
                        br=1;
		        if(dmode == DMODE_FLIGHT)
               		    event.key.keysym.sym=SDLK_RETURN; //+ throttle
			else
			    event.key.keysym.sym=SDLK_r; //radar mapper

            		Keymap_KeyDown(&event.key.keysym);
			break;

                case SDLK_LSHIFT: //Y
		        if(dmode == DMODE_FLIGHT)
             		    event.key.keysym.sym=SDLK_e; //ecm
			else
             		    event.key.keysym.sym=SDLK_F8; //hyperspace

            		Keymap_KeyDown(&event.key.keysym);
			break;

           	case SDLK_SPACE: //X
		        if(dmode == DMODE_FLIGHT)
             		    event.key.keysym.sym=SDLK_F1; //views
			else
             		    event.key.keysym.sym=SDLK_x; //eject

             		Keymap_KeyDown(&event.key.keysym);
             		break;

	   	case SDLK_RETURN: //START
                        bs=1;
            		event.key.keysym.sym=SDLK_F7; //motor
             		Keymap_KeyDown(&event.key.keysym);

             		event.key.keysym.sym=SDLK_p; //claves
             		Keymap_KeyDown(&event.key.keysym);
             		break;

		case SDLK_ESCAPE: //SELECT
			if((br)&&(bl))
                        {
       			    bQuitProgram = TRUE;
       			    SDL_Quit();
       			    exit(0);
                        }
                        else if (bs)
             		    Keymap_KeyDown(&event.key.keysym);
			else
                        {
             		    dmode++; //change mode
                            if(dmode > 2)
                               dmode = 0;
                            switch(dmode)
                            {
				case DMODE_CURSOR:
			    	    input.abs_x = ratonx = savedrx; //restore cursor
			    	    input.abs_y = ratony = savedry;
				    mouse_bmp = cursor_bmp; //change atari cursor (hostcall.c)
                                    incraton = 2;
                                    break;
				case DMODE_FLIGHT:
                                    savedrx = ratonx; //save cursor pos
                                    savedry = ratony;
			    	    input.abs_x = ratonx = 298;
			    	    input.abs_y = ratony = 150;
				    //mouse_bmp = stick_bmp; //right button not show cursor :(
             		            Input_MousePress (SDL_BUTTON_RIGHT);
                                    raton = 1;
                                    incraton = 3;
                                    break;
				case DMODE_MAP:
                                    raton = 0;
				    mouse_bmp = world_bmp;
             		            Input_MouseRelease (SDL_BUTTON_RIGHT);
                                    break;
                            }
                            //printf("Debug: raton cambiado a modo: %i\n", raton);
                        }

			break;
        }
        break;


    case SDL_KEYUP:
	switch(event.key.keysym.sym)
        {
		case SDLK_RIGHT:
             		pr=0;
		        if (dmode == DMODE_MAP)
             		    Keymap_KeyUp(&event.key.keysym);
	     		break;

		case SDLK_LEFT:
             		pl=0;
		        if (dmode == DMODE_MAP)
             		    Keymap_KeyUp(&event.key.keysym);
	     		break;

		case SDLK_UP:
                        pu=0;
		        if (dmode == DMODE_MAP)
             		    Keymap_KeyUp(&event.key.keysym);
	     		break;

		case SDLK_DOWN:
                        pd=0;
		        if (dmode == DMODE_MAP)
             		    Keymap_KeyUp(&event.key.keysym);
	     		break;

           	case SDLK_LCTRL: //A
/*		        if(dmode == DMODE_FLIGHT)
                        {
             		    //event.key.keysym.sym=SDLK_SPACE; // fire laser
             		    //Keymap_KeyUp(&event.key.keysym);
             		    Input_MouseRelease (SDL_BUTTON_LEFT);
                        }
                        else
*/
             		    Input_MouseRelease (SDL_BUTTON_LEFT);
             		break;

		case SDLK_LALT:  //B
          		//event.key.keysym.sym=SDLK_m; // fire misile
		        if(dmode == DMODE_FLIGHT)
                        {
             		    event.key.keysym.sym=SDLK_F3; //status ship
             		    Keymap_KeyUp(&event.key.keysym);
                        }
                        else
             		    Input_MouseRelease (SDL_BUTTON_RIGHT);

                        raton = 0;
			break;

                case SDLK_TAB: //L
                        bl=0;
		        if(dmode == DMODE_FLIGHT)
             		    event.key.keysym.sym=SDLK_RSHIFT; //- throttle
			else
             		    event.key.keysym.sym=SDLK_h; // hyperspace cloud analyser ???

             		Keymap_KeyUp(&event.key.keysym);
                        break;

		case SDLK_BACKSPACE: //R
                        br=0;
		        if(dmode == DMODE_FLIGHT)
               		    event.key.keysym.sym=SDLK_RETURN; //+ throttle
			else
			    event.key.keysym.sym=SDLK_r; //radar mapper

             		Keymap_KeyUp(&event.key.keysym);
			break;

                case SDLK_LSHIFT: //Y
		        if(dmode == DMODE_FLIGHT)
             		    event.key.keysym.sym=SDLK_e; //ecm
			else
             		    event.key.keysym.sym=SDLK_F8; //hypespace

             		Keymap_KeyUp(&event.key.keysym);
			break;

           	case SDLK_SPACE: //X
		        if(dmode == DMODE_FLIGHT)
             		    event.key.keysym.sym=SDLK_F1; //views
			else
             		    event.key.keysym.sym=SDLK_x; //eject
             		Keymap_KeyUp(&event.key.keysym);
             		break;

	   	case SDLK_RETURN: //START
                        bs=0;
            		event.key.keysym.sym=SDLK_F7; //motor
             		Keymap_KeyUp(&event.key.keysym);

             		event.key.keysym.sym=SDLK_p; //claves
             		Keymap_KeyUp(&event.key.keysym);
             		break;

		case SDLK_ESCAPE: //SELECT
             		Keymap_KeyUp(&event.key.keysym);

        }
        break;

   }
Input_Update ();
}


/*-----------------------------------------------------------------------*/
/*
  Check for any passed parameters
*/
void Main_ReadParameters(int argc, char *argv[])
{
  int i;

  /* Scan for any which we can use */
  for(i=1; i<argc; i++)
  {
    if (strlen(argv[i])>0)
    {
      if (!strcmp(argv[i],"--help") || !strcmp(argv[i],"-h"))
      {
        printf("Usage:\n frontier [options]\n"
               "Where options are:\n"
               "  --help or -h          Print this help text and exit.\n"
               "  --fullscreen or -f    Try to use fullscreen mode.\n"
               "  --nosound             Disable sound (faster!).\n"
              );
        exit(0);
      }
      else if (!strcmp(argv[i],"--fullscreen") || !strcmp(argv[i],"-f"))
      {
        bUseFullscreen=TRUE;
      }
      else if ( !strcmp(argv[i],"--nosound") )
      {
        bDisableSound=TRUE;
        ConfigureParams.Sound.bEnableSound = FALSE;
      }
      else
      {
	      /* some time make it possible to read alternative
	       * names for fe2.bin from command line */
	      fprintf(stderr,"Illegal parameter: %s\n",argv[i]);
      }
    }
  }
}


/*-----------------------------------------------------------------------*/
/*
  Initialise emulation
*/
void Main_Init(void)
{
  /* Init SDL's video subsystem. Note: Audio and joystick subsystems
     will be initialized later (failures there are not fatal). */
  if(SDL_Init(SDL_INIT_VIDEO | SDL_INIT_TIMER | SDL_INIT_JOYSTICK) < 0)
  {
    fprintf(stderr, "Could not initialize the SDL library:\n %s\n", SDL_GetError() );
    exit(-1);
  }
  
  SDL_JoystickOpen(0);

  Misc_SeedRandom(1043618);
  SDLGui_Init();
  Screen_Init();
  Init680x0();                  /* Init CPU emulation */
  Audio_Init();
  Keymap_Init();

  GemDOS_Init();
  GemDOS_InitDrives();

  if(bQuitProgram)
  {
    SDL_Quit();
    exit(-2);
  }
 
}


/*-----------------------------------------------------------------------*/
/*
  Un-Initialise emulation
*/
void Main_UnInit(void)
{
  Screen_ReturnFromFullScreen();
  GemDOS_UnInitDrives();
  Audio_UnInit();
  SDLGui_UnInit();
  Screen_UnInit();

  /* SDL uninit: */
  SDL_Quit();

}

static Uint32 vbl_callback ()
{
	FlagException (0);
	return 20;
}

void sig_handler (int signum)
{
	if (signum == SIGSEGV) {
		printf ("Segfault! All is lost! Abandon ship!\n");
		Call_DumpDebug ();
		abort ();
	}
}

/*-----------------------------------------------------------------------*/
/*
  Main
*/
int main(int argc, char *argv[])
{
  struct sigaction act;

  act.sa_handler = sig_handler;
  act.sa_flags = 0;
  sigaction (SIGSEGV, &act, NULL);
	
  /* Generate random seed */
  srand( time(NULL) );

  /* Get working directory, if in MSDev force */
  Misc_FindWorkingDirectory(argv[0]);
#ifdef FORCE_WORKING_DIR
  getcwd(szWorkingDir, MAX_FILENAME_LENGTH);
#endif

  /* Set default configuration values: */
  Configuration_SetDefault();

  /* Check for any passed parameters */
  Main_ReadParameters(argc, argv);



  //fd = open( "/dev/touchscreen/wm97xx", O_RDONLY | O_NOCTTY );	// Lectura del tactil

  //if (fd){


	  /* Init emulator system */
	  Main_Init();

	  /* Switch immediately to fullscreen if user wants to */
	  if( bUseFullscreen )
	    Screen_EnterFullScreen();

	  SDL_AddTimer (20, &vbl_callback, NULL);
  
	  /* Run emulation */
	  Main_UnPauseEmulation();
	  Start680x0();                 /* Start emulation */

	  /* Un-init emulation system */
	  Main_UnInit();
 //}
  return(0);
}


